import logging
import os
import subprocess
import tempfile
from multiprocessing import Process
from typing import Text

import pkg_resources
from sanic import Blueprint, response
from sanic.exceptions import NotFound
from sanic.request import Request
from sanic.response import text

import rasax.community
from rasa.cli.utils import print_success, print_info, print_error
from rasax.community import utils
from rasax.community.constants import API_URL_PREFIX

logger = logging.getLogger(__name__)

PACKAGE_INTERFACE_DIRECTORY = pkg_resources.resource_filename(
    rasax.community.__name__, "interface")

# project root, I know, lots of .. :D
root_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                        "..", "..", "..", "..", "..", "..")


def build_interface() -> Text:
    """Compile the frontend using the repos make file and move it to a temp dir.

    Returns the temporary directory containing the compiled frontend."""

    tmpdir = str(tempfile.mkdtemp())
    frontend_directory = os.path.join(tmpdir, "build")

    def _write_build_in_progress_index_html():
        os.mkdir(frontend_directory)
        utils.write_bytes_to_file(
            os.path.join(frontend_directory, "index.html"),
            '<html><head><meta http-equiv="refresh" content="1"></head>'
            '<body><img src="https://cataas.com/cat/says/Frontend compiling - '
            'and now...we wait" alt="Frontend is compiling..."/><body></html>'
            ''.encode("utf-8"))

    def run_build(cwd, output):
        def write_failure_index_html():
            utils.write_bytes_to_file(
                os.path.join(frontend_directory, "index.html"),
                "Frontend build failed! Check the logs for details.".encode("utf-8"))

        print_info("Building frontend (development mode)...")
        # this will always use the frontend enterprise build, as in this case we
        # have the source anyways (won't happen in packaged build)
        if subprocess.call(["make", "install-frontend"], cwd=cwd):
            print_error("Failed to install frontend dependencies. "
                        "Check logs for details.")
            write_failure_index_html()
        elif subprocess.call(["make", "build-frontend-enterprise"], cwd=cwd):
            print_error("Failed to build frontend code. Check logs for details.")
            write_failure_index_html()
        elif subprocess.call(["cp", "-r", "src/frontend/build", output], cwd=cwd):
            print_error("Failed to copy frontend to package. Check logs for details.")
            write_failure_index_html()
        else:
            print_success("Finished building frontend, serving from {}."
                          "".format(output))

    p = Process(target=run_build, args=(root_dir, tmpdir))
    p.start()

    _write_build_in_progress_index_html()

    return frontend_directory


def locate_interface() -> Text:
    """Check if there is a packaged interface - if not build it from source.

    Returns the path to the interface directory."""

    if utils.is_enterprise_installed():
        from rasax.enterprise.interface import \
            PACKAGE_ENTERPRISE_INTERFACE_DIRECTORY
        pkg_base = PACKAGE_ENTERPRISE_INTERFACE_DIRECTORY
    else:
        pkg_base = PACKAGE_INTERFACE_DIRECTORY

    pkg_index = os.path.join(pkg_base, "index.html")
    if os.path.exists(pkg_index):
        return pkg_base
    else:
        if not os.environ.get("SKIP_FRONTEND_BUILD"):
            return build_interface()
        else:
            external_frontend = os.path.join(root_dir, "src", "frontend", "build")
            print("Using external frontend build - make sure there is a frontend "
                  "available in {}".format(os.path.abspath(external_frontend)))
            return external_frontend


def blueprint() -> Blueprint:
    """Serve the Rasa X interface."""

    interface_directory = locate_interface()

    index_html = os.path.join(interface_directory, "index.html")

    interface = Blueprint('interface')
    interface.static('/static', os.path.join(interface_directory,
                                             "static"))
    interface.static('/', index_html)

    @interface.exception(NotFound)
    async def ignore_404s(request: Request, exception):
        if request.path.startswith(API_URL_PREFIX):
            # tries to avoid answering any API calls with
            # the frontend response
            return text("Not found.", status=404)
        logger.debug("Answered 404 with index.html."
                     "Request url '/': '{}'".format(request.url))
        # we need this, as the frontend does its own routing, but needs
        # the server to always deliver the index.html to not interfere
        return await response.file(index_html)

    return interface
